//	general header for all functions in directory

#ifndef HEADER_FILE
#define HEADER_FILE

#include <string>
using namespace std;

//	----------	vdop.cpp	--------------

void vdcopy(const int n, double xp[], const int ix,
                   const double yp[], const int iy);

void vdplus(const int n, double x[], const int ix,
                   const double y[], const int iy,
		   const double z[], const int iz);

void vddiv(const int n,  double x[], const int ix,
                   const double y[], const int iy,
		   const double z[], const int iz);

void vdmult(const int n, double x[], const int ix,
                   const double y[], const int iy,
		   const double z[], const int iz);

double dscapr(const int n, const double x[], const double y[]);

void vdaxpy(const int n, double x[], const double y[],
                   const double& a,  const double z[] );

void vicopy(const int n, int xp[], const int ix,
                   const int yp[], const int iy);

//	----------	vdop.cpp	--------------
/**
 * The quadratic domain [0,1] x [0,1] is divided into @p procx * @p procy subdomains
   numbered rowise.
   According to my process with rank number @p myid, the vector @p neigh contains the ranks
   of the neighbours t the South, East, North, West.
 * @param[in] myid	my rank
 * @param[in] procx	number of processes in x-direction
 * @param[in] procy	number of processes in y-direction
 * @param[out] neigh	vector of 4 elements containing the ranks of neighbouring processes, -1 indicates a boundary in this direction
 * @param[out] color	red/black coloring of subdomains (0/1)
*/
void IniGeom(const int myid, const int procx, const int procy, int neigh[], int& color);


/**
 * The quadratic domain [0,1] x [0,1] is divided into @p procx * @p procy subdomains
   numbered rowise.
   According to my process with rank number @p myid, the coordinates of the lower
   left corner (@p xl, @p yb) and of the upper right corner (@p xr, @p yt) are generated.
 * @param[in] myid	my rank
 * @param[in] procx	number of processes in x-direction
 * @param[in] procy	number of processes in y-direction
 * @param[out] xl	x-coordinate of left boundary
 * @param[out] xr	x-coordinate of right boundary
 * @param[out] yb	y-coordinate of lower boundary
 * @param[out] yt	y-coordinate of upper boundary
*/
void IniCoord(const int myid, const int procx, const int procy,
	      double& xl, double& xr, double& yb, double& yt);


/**
 * Computes @p w := @p w + @p s  at the interface/boundary nodes on the
 * boundary @p ib .  South (ib==1), East (ib==2), North (ib==3), West (ib==4)
 * @param[in] ib	my local boundary
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in,out] w	vector for all nodes of local discretization
 * @param[in] s		short vector with values on boundary @p ib
*/
void AddBound(const int ib, const int nx, const int ny, double w[], const double s[]);


//	----------	getbound.cpp	--------------
/**
 * Copies the values of @p w corresponding to boundary @p ib
 * onto vector s.  South (ib==1), East (ib==2), North (ib==3), West (ib==4).
 * The vector @p s has to be long enough!!
 * @param[in] ib	my local boundary
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] w		vector for all nodes of local discretization
 * @param[out] s	short vector with values on boundary @p ib
*/
// GH_NOTE: Absicherung bei s !!
void GetBound(const int ib, const int nx, const int ny, const double w[], double s[]);


//	----------	skalar.cpp	--------------
/*
	Compute scalar product of two vectors
	
	IN:	n	length of the vectors
		x,y 	vectors
	OUT:	scalar	scalar product
*/
/**
 * Computes the parallel scalar product of two double vectors. 
 * One vecctor has to be accumulated the other one has to be distributed.
 * @param[in] n		length of the vectors
 * @param[in] x		vector
 * @param[in] y		vector
 * @see VecAccu
*/
// double skalar(const int n, const double x[], const double y[], const MPI::Intracomm& icomm);

//	----------	userset.cpp	--------------
/**
 * User function: assignes values to solution vector @p u.
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[out] u	vector for all nodes of local discretization
*/
void SetU(const int nx, const int ny, double u[]);

/**
 * User function: assignes values to right hand side vector @p f.
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[out] f	vector for all nodes of local discretization
*/
void SetF(const int nx, const int ny, double f[]);

/**
 * User function: f(@p x,@p y) 
 * @param[in] x		x-coordinate of discretization point
 * @param[in] y		y-coordinate of discretization point
 * @return  value for right hand side f(@p x,@p y)
 */
double FunctF(const double& x, const double& y);

/**
 * User function: u(@p x,@p y) 
 * @param[in] x		x-coordinate of discretization point
 * @param[in] y		y-coordinate of discretization point
 * @return  value for solution vector u(@p x,@p y)
 */
double FunctU(const double& x, const double& y);

//	----------	savevekp.cpp	--------------

/**
 * Stores the values of vector @p u of subdomain @p myid into a file.
 * The file stores rowise the x- and y- coordinates together with the value from  @p u .
 * The domain [@p xl, @p xr] x [@p yb, @p yt] is discretized into @p nx x @p ny intervals.
 * @param[in] myid	my rank
 * @param[in] name	basename of file name (file name will be extended by the rank number)
 * @param[in] u		local vector
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] xl	x-coordinate of left boundary
 * @param[in] xr	x-coordinate of right boundary
 * @param[in] yb	y-coordinate of lower boundary
 * @param[in] yt	y-coordinate of upper boundary
*/
void SaveVectorP(const int myid, const string& name, const double u[], 
		 const int nx, const int ny, 
		 const double& xl, const double& xr, const double& yb, const double& yt);

//	----------	applybc.cpp	--------------
/**
 * Applies Dirichlet boundary conditions to symmetric CSR matrix  (@p sk,@p id, @p ik) and right hand side @p f 
 * of a system of equations resulting from a FEM/FDM discretization
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] neigh	vector of 4 elements containing the ranks of neighbouring processes, -1 indicates a boundary in this direction
 * @param[in] u		local vector with appropriately initialized b.c.
 * @param[in,out] sk	vector non-zero entries of CSR matrix
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[in,out] f	right hand side
*/
void ApplyDirichletBC(const int nx, const int ny, const int neigh[], 
	const double u[], double sk[], const int id[], const int* , double f[]);

//	----------	crsmult.cpp	--------------
/**
 * Multiplies vector @p u with a symmetric CSR matrix K(@p sk,@p id, @p ik). Operation can be performed 
 * for only a subset [@p iza, @p ize] of matrix rows.
 * w := w + alpha*K*u
 * @param[in] iza	first matrix row (starting at 0)
 * @param[in] ize	last matrix row
 * @param[in,out] w	distributed local vector.
 * @param[in] u		accumulated local vector.
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[in] sk	vector non-zero entries of CSR matrix
 * @param[in] alfa	scaling factor
*/
void  CrsMult(const int iza, const int ize, double w[], const double u[], 
	const int id[], const int ik[], const double sk[], const double& alfa);


/**
 * Copies the diagonal entries of a symmetric CSR matrix (@p sk,@p id, @p ik)
 * onto a vector @p d.
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] sk	vector non-zero entries of CSR matrix
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[out] d	accumulated diagonal entries of CSR matrix
*/

void GetDiag(const int nx, const int ny, const double sk[], const int id[], double d[]);

//	----------	jacsolve.cpp	--------------

/**
 * Solves linear system of equations  K @p u = @p f  via the Jacobi iteration.
 * We use a distributed symmetric  CSR matrix K(@p sk,@p id, @p ik) and initial guess of the 
 * solution is set to 0.
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] sk	vector non-zero entries of CSR matrix
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[in] f		distributed local vector storing the right hand side
 * @param[out] u	accumulated local vector storing the solution.
*/

void JacobiSolve(const int nx, const int ny, 
	const double sk[], const int id[], const int ik[], 
	const double f[], double u[] );


//	----------	getmatrix.cpp	--------------
/**
 * Determines the coordinates of the dicretization nodes of the domain [@p xl, @p xr] x [@p yb, @p yt] 
 * which is discretized into @p nx x @p ny intervals.
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] xl	x-coordinate of left boundary
 * @param[in] xr	x-coordinate of right boundary
 * @param[in] yb	y-coordinate of lower boundary
 * @param[in] yt	y-coordinate of upper boundary
 * @param[out] xc	coordinate vector of length 2n with x(2*k,2*k+1) as coodinates of node k
*/

void GetCoordsInRectangle(const int nx, const int ny, 
	const double& xl, const double& xr, const double& yb, const double& yt, 
	double xc[]);
/**
 * Determines the element connectivity of linear triangular elements of a FEM discretization  
 * of a rectangle using @p nx x @p ny equidistant intervals for discretization.
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[out] ia	element connectivity matrix with ia(3*s,3*s+1,3*s+2) as node numbers od element s
*/
void GetConnectivityInRectangle(const int nx, const int ny, int ia[]);

/**
 * Determines the sparsity pattern (@p id, @p ik) of a symmetric sparse CSR matrix assuming 
 * linear triangular elements in a rectangle using @p nx x @p ny equidistant intervals for discretization.
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[out] id	index vector containing the first entry in a CSR row
 * @param[out] ik	column index vector of CSR matrix
*/
void GetMatrixpointersInRectangle(const int nx, const int ny, int id[], int ik[]);

/**
 * Calculates the element stiffness matrix @p ske and the element load vector @p fe 
 * of one triangular element with linear shape functions.
 * @param[in]	ia	node indices of the three element vertices
 * @param[in]	xc	vector of node coordinates with x(2*k,2*k+1) as coodinates of node k
 * @param[out] ske	element stiffness matrix
 * @param[out] fe	element load vector
 */ 
void CalcElem(const int ial[3], const double xc[], double ske[3][3], double fe[3]);

/**
 * Adds the element stiffness matrix @p ske and the element load vector @p fe 
 * of one triangular element with linear shape functions to the appropriate positions in 
 * the symmetric stiffness matrix, stored as CSR matrix K(@p sk,@p id, @p ik)
 * @param[in] ia	node indices of the three element vertices
 * @param[in] ske	element stiffness matrix
 * @param[in] fe	element load vector
 * @param[out] sk	vector non-zero entries of CSR matrix
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[out] f	distributed local vector storing the right hand side
*/
void AddElem(const int ial[3], const double ske[3][3], const double fe[3], 
	double sk[], const int id[], const int ik[], double f[]);

/**
 * Print entries of the CSR matrix K(@p sk,@p id, @p ik).
 * @param[in] nnode	number of rows
 * @param[in] sk	vector non-zero entries of CSR matrix
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
*/
void DebugMatrix(const int nnode, const double sk[], const int id[], const int ik[]);

/**
 * Computes the stiffness matrix  K(@p sk,@p id, @p ik) and load vector @p f for the Poisson problem 
 * without incorporating the boundary conditions
 * in a domain [@p xl, @p xr] x [@p yb, @p yt] which is discretized into @p nx x @p ny intervals.
 * The matrix pattern is generated and the matrix is stored as CSR matrix(@p sk,@p id, @p ik).
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] xl	x-coordinate of left boundary
 * @param[in] xr	x-coordinate of right boundary
 * @param[in] yb	y-coordinate of lower boundary
 * @param[in] yt	y-coordinate of upper boundary
 * @param[out] sk	vector non-zero entries of CSR matrix
 * @param[out] id	index vector containing the first entry in a CSR row
 * @param[out] ik	column index vector of CSR matrix
 * @param[out] f	distributed local vector storing the right hand side
 */
void GetMatrix(const int nx, const int ny, 
		const double& xl, const double& xr, const double& yb, const double& yt, 
		double sk[], int id[], int ik[], double f[]);

#endif
