#ifndef MG_HEADER_FILE
#define MG_HEADER_FILE

#define MAXLEVEL 20
/* improved coarse grid solver */
/* this structure stores all pointers to vectors, matrices of all elevels */

struct PointerStruct
{ int		nx[MAXLEVEL],		/* intervalls in x/y direction */
		ny[MAXLEVEL];
				/* pointers for array .. on all levels */
  int		*id[MAXLEVEL],		/* row end pointer       */
  		*ik[MAXLEVEL];		/* column index          */
  double	*sk[MAXLEVEL],		/* matrix entries        */
  		*u[MAXLEVEL],		/* solution/correction   */
		*f[MAXLEVEL],		/* r.h.d/defect          */
		*dd[MAXLEVEL],		/* accumulated diagonal  */
		*aux[MAXLEVEL];		/* aux. vector (d,w,...) */
/*  local to global numbering on coarse grid */
  int		*loc2glob;
/*  accumulated coarse matrix */
  int		nglob, *cid, *cik;
  double	*csk;
};


/* this structure stores control parameters of the multigrid algorithm */

struct ControlStruct
{ int		cycle,			/* kind of mg cycle (1/2) */
		npre, npost;      	/* number of pre/postsmoothing sweeps */
};

// ------------------------------ applybc1.cpp  ---------------------------------------------
/**
 * Applies Dirichlet boundary conditions to symmetric CSR matrix  (@p sk,@p id, @p ik) and right hand side @p f 
 * of a system of equations resulting from a FEM/FDM discretization.
 * This is a special version function ApplyDirichletBC for parallel matrix generation taking into account 
 * distributed matrix and distributed rhs.
 * 
 * @see ApplyDirichletBC
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] neigh	vector of 4 elements containing the ranks of neighbouring processes, -1 indicates a boundary in this direction
 * @param[in] u		local vector with appropriately initialized b.c.
 * @param[in,out] sk	vector non-zero entries of CSR matrix
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[in,out] f	right hand side
*/
void ApplyDirichletBC1(const int nx, const int ny, const int neigh[], 
	const double u[], double sk[], const int id[], const int ik[] , double f[]);

/**
 * Determines for a symmetrix CSR matrix the index of element (@p row, @p col).
 * 
 * @param[in] row	row index
 * @param[in] col 	colum index
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @return index of element (@p row, @p col). -1 indicates that this entry is not available.
 */
int AdrCRS(const int row, const int col, const int id[], const int ik[], const int nnode);

// ------------------------------  allocati.cpp ---------------------------------------------
/**
 * Allocates the memory for @p nlevels levels for a rectangle discretized 
 * into @p nx x @p ny equidistant intervals on the coarsest grid at level 0. 
 * The geometric multigrid hierarchy assume bisection of intervals as refinement strategy.
 * 
 * @param[in] nlevels	number of multigrid levels
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction 
 * @param[out] *pp	pointer to the structure performing the grid management
 * @return -1 indicates an error.
 */ 

int AllocLevels(const int nlevels, const int nx, const int ny, PointerStruct* pp);

/**
 * Inializes the stiffness matrix, right hand side, and boundary conditions for the 
 * discretization on level @p nlevels in the domain [@p xl, @p xr] x [@p yb, @p yt].
 *
 * memory for @p nlevels levels for a rectangle discretized 
 * into @p nx x @p ny equidistant intervals on the coarsest grid at level 0. 
 * The geometric multigrid hierarchy assume bisection of intervals as refinement strategy.
 * 
 * @param[in] xl	x-coordinate of left boundary
 * @param[in] xr	x-coordinate of right boundary
 * @param[in] yb	y-coordinate of lower boundary
 * @param[in] yt	y-coordinate of upper boundary
 * @param[in] neigh	vector of 4 elements containing the ranks of neighbouring processes, -1 indicates a boundary in this direction
 * @param[in] nlevels	number of multigrid levels
 * @param[in] *pp	pointer to the structure performing the grid management
 * @return -1 indicates an error.
 */
int IniLevels(const double xl, const double xr, const double yb, const double yt, 
	      const int neigh[],
	      const int nlevels, const PointerStruct *pp);

/**
 * Frees the memory allocated for the multigrid hierarchy.
 * 
 * @param[in] nlevels	number of multigrid levels
 * @param[out] *pp	pointer to the structure performing the grid management
 * @return -1 indicates an error.
 * @see AllocLevels
 */
int FreeLevels(const int nlevels, PointerStruct* pp);

// -------------------------------  gridtransfer.cpp --------------------------------------------
/**
 * Bilinear interpolation of a coarse grid vector @p w_c onto a fine grid vector @p w.
 * The underlying discretization is a tensor product grid.
 * 
 * @param[in] nx	number of fine grid intervals in x-direction
 * @param[in] ny	number of fine grid intervals in y-direction 
 * @param[out] w	interpolated vector on fine grid
 * @param[in] nx_c	number of fine grid intervals in x-direction
 * @param[in] ny_c	number of fine grid intervals in y-direction
 * @param[in] w_c	original vector on coarse grid
 * @param[in] neigh	vector of 4 elements containing the ranks of neighbouring processes, -1 indicates a Dirichlet boundary in this direction. Use other negative values for Neumann boundary
 */
void Interpolate(const int nx, const int ny, double w[],  
		  const int nx_c, const int ny_c, const double w_c[], const int neigh[]);

/**
 * Bilinear restriction of a fine grid vector @p w onto a coarse grid vector @p w_c.
 * The underlying discretization is a tensor product grid.
 * 
 * @param[in] nx	number of fine grid intervals in x-direction
 * @param[in] ny	number of fine grid intervals in y-direction
 * @param[in] w	interpolated vector on fine grid
 * @param[in] nx_c	number of fine grid intervals in x-direction
 * @param[in] ny_c	number of fine grid intervals in y-direction
 * @param[out] w_c	original vector on coarse grid
 * @param[in] neigh	vector of 4 elements containing the ranks of neighbouring processes, -1 indicates a Dirichlet boundary in this direction. Use other negative values for Neumann boundary
 * @see Interpolate
 */
void Restrict(const int nx, const int ny, const double w[],  
	      const int nx_c, const int ny_c, double w_c[], const int neigh[]);

// -----------------------------  jacsmooth.cpp  ----------------------------------------------
/**
 * Iterates @p maxiter with the linear system of equations  K @p u = @p f  via the Jacobi iteration.
 * We use a distributed symmetric  CSR matrix K(@p sk,@p id, @p ik).
 * 
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] sk	vector non-zero entries of CSR matrix
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[in] f		distributed local vector storing the right hand side
 * @param[in,out] u	accumulated local vector storing the solution.
 * @param[in] dd	accumualted diagonal of striffness matrix
 * @param[out] aux	auxilliary vector used for the intermediate residuum in the iteration
 * @param[in] maxiter	number of iterations / smoothing weeps
*/
void JacobiSmooth(const int nx, const int ny,
		  const double sk[], const int id[], const int ik[],
		  const double f[], double u[], const double dd[],
		  double aux[], const int maxiter );

/**
 * Solves the global parallel system with Jacobi iterations until a predefined accuracy is reached.
 * We use a distributed symmetric  CSR matrix K(@p sk,@p id, @p ik).
 *
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] sk	vector non-zero entries of CSR matrix
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[in] f		distributed local vector storing the right hand side
 * @param[in,out] u	accumulated local vector storing the solution.
 * @param[in] dd	accumualted diagonal of striffness matrix
 * @param[out] aux	auxilliary vector used for the intermediate residuum in the iteration
*/
void JacobiSolve2(const int nx, const int ny,
		  const double sk[], const int id[], const int ik[],
		  const double f[], double u[], const double dd[], double aux[]);
// -----------------------------  mgm.cpp  ----------------------------------------------

void MGM(const PointerStruct* const pp, const ControlStruct* const crtl,
	 const int level, const int neigh[]
// 	 ,const int color, const int myid, const MPI::Intracomm& icomm
	);


void MGMSolver(const PointerStruct* const pp, const ControlStruct* const crtl, const int level, const  int neigh[]
// 	       , const int color, const int myid, const MPI::Intracomm& icomm
	      );


// ----------------------------- ----------------------------------------------

#endif
